#!/usr/bin/env python3
"""
Pure python implementation of the gimp gtp tool preset format
"""
from __future__ import annotations
import argparse
import brackettree

class ParenFileValue:
	"""
	A parentheses-based file format
	(possibly "scheme" language?)
	"""
	def __init__(self, name=None, value="", children=None):
		self.name = name
		try:
			float(value)
			self.value = value
		except ValueError:
			self.value = "\"" + value + "\""
		if value in ("yes", "no"):
			self.value = value
		self.children = children

	def _addValue(self, bufArray):
		if self.name is None: # first value is the name
			self.name = bufArray
		if bufArray:
			bufArray = ''.join(bufArray)
			if bufArray in ('yes', 'no'): # boolean
				self.value.append(bufArray == 'yes')
			elif bufArray[0].isdigit():
				self.value.append(float(bufArray))
			elif bufArray[0] == '"':
				self.value.append(bufArray[1:-1])
			else:
				raise Exception('What kind of value is "' + bufArray + '"?')

	def __repr__(self):
		""" Get a textual representation of this object """
		ret = []
		ret.append('(' + str(self.name))
		ret.append(" " + self.value)
		if self.children is None:
			ret.append(")")
		ret.append("\n")
		if self.children is not None:
			for index, child in enumerate(self.children):
				ret.append("    (" + str(child.name))
				ret.append(" " + child.value + ")")
				if index == len(self.children) - 1:
					ret.append(")")
				ret.append("\n")
		return "".join(ret)


def parenFileDecode(data):
	"""
	Decode a parentheses-based file format
	(possibly "scheme" language?)
	"""
	nodes = brackettree.Node(data)
	values = walkTree(nodes.items)
	return values



def walkTree(items):
	""" walk the tree """
	values = []
	for item in items:
		if isinstance(item, brackettree.RoundNode):
			if len(item.items) == 1:
				values.append(ParenFileValue(item.items[0].split(" ")[0].strip(),
				item.items[0].split(" ")[1].strip()))
			if len(item.items) == 2:
				values.append(ParenFileValue(item.items[0].strip(),
				item.items[1].items[0].strip()))
			elif len(item.items) > 2:
				values.append(ParenFileValue(item.items[0].strip(),
				item.items[1].items[0].strip(),
				walkTree(item.items[2:])))
	return values


def parenFileEncode(values):
	"""
	encode a values tree to a buffer
	"""
	ret = []
	ret.append('# GIMP tool preset file\n\n')
	ret.append('')
	for val in values:
		if val.name is not None:
			ret.append(str(val))
	ret.append('')
	ret.append('\n# end of GIMP tool preset file\n')
	ret.append('')
	return "".join(ret)


class GimpGtpToolPreset:
	"""
	Pure python implementation of the gimp gtp tool preset format
	"""
	def __init__(self, fileName=None):
		self.fileName = None
		self.values = []
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: Union[BytesIO, str]):
		"""
		load a gimp file

		:param fileName: can be a file name or a file-like object
		"""
		if hasattr(fileName, 'read'):
			self.fileName = fileName.name
			f = fileName
		else:
			self.fileName = fileName
			f = open(fileName, 'r')
		data = f.read()
		self.decode(data)
		f.close()

	def decode(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		"""
		self.values = parenFileDecode(data)
		return index

	def encode(self):
		"""
		encode to a byte array
		"""
		return parenFileEncode(self.values).encode('utf-8')

	def save(self, tofileName=None, toExtension=None):
		"""
		save this gimp tool preset to a file
		"""
		if toExtension is None:
			if tofileName is not None:
				toExtension = tofileName.rsplit('.', 1)
				if len(toExtension) > 1:
					toExtension = toExtension[-1]
				else:
					toExtension = None
		if not hasattr(tofileName, 'write'):
			f = open(tofileName, 'wb')
		f.write(self.encode())
		f.close()

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		for v in self.values:
			ret.append(v.__repr__(indent + '\t'))
		return '\n'.join(ret)


if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("gimpGtpToolPreset.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	parser.add_argument("--dump", action="store_true",
	help="dump info about this file")
	args = parser.parse_args()

	gimpGtpToolPreset = GimpGtpToolPreset(args.xcfdocument)

	if args.dump:
		print(gimpGtpToolPreset)



#!/usr/bin/env python3
"""
Stuff related to vectors/paths within a gimp document
"""
#from .GimpIOBase import GimpIOBase
from __future__ import annotations
from binaryiotools import IO
from .GimpParasites import GimpParasite

class GimpVector:
	"""
	A gimp brush stroke vector
	"""
	def __init__(self, parent):
		#GimpIOBase.__init__(self, parent)
		self.name = ''
		self.uniqueId = 0
		self.visible = True
		self.linked = False
		self.parasites = []
		self.strokes = []

	def decode(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		ioBuf = IO(data, index, boolSize=32)
		self.name = ioBuf.sz754
		self.uniqueId = ioBuf.u32
		self.visible = ioBuf.bool
		self.linked = ioBuf.bool
		numParasites = ioBuf.u32
		numStrokes = ioBuf.u32
		for _ in range(numParasites):
			p = GimpParasite()
			ioBuf.index = p.decode(ioBuf.data, ioBuf.index)
			self.parasites.append(p)
		for _ in range(numStrokes):
			gs = GimpStroke(self)
			ioBuf.index = gs.decode(ioBuf.data, ioBuf.index)
			self.strokes.append(p)
		return ioBuf.index

	def encode(self):
		"""
		encode to binary data
		"""
		ioBuf = IO(boolSize=32)
		ioBuf.sz754 = self.name
		ioBuf.u32 = self.uniqueId
		ioBuf.bool = self.visible
		ioBuf.bool = self.linked
		ioBuf.u32 = len(self.parasites)
		ioBuf.u32 = len(self.strokes)
		for p in self.parasites:
			ioBuf.addBytes(p.encode())
		for gs in self.strokes:
			ioBuf.addBytes(gs.encode())
		return ioBuf.data

	def __repr__(self, indent: str='') -> str:
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Name: ' + str(self.name))
		ret.append('Unique ID (tattoo): ' + str(self.uniqueId))
		ret.append('Visible: ' + str(self.visible))
		ret.append('Linked: ' + str(self.linked))
		if self.parasites:
			ret.append('Parasites: ')
			for item in self.parasites:
				ret.append(item.__repr__(indent + '\t'))
		if self.strokes:
			ret.append('Strokes: ')
			for item in self.strokes:
				ret.append(item.__repr__(indent + '\t'))
		return indent + (('\n' + indent).join(ret))


class GimpStroke:
	"""
	A single stroke within a vector
	"""

	STROKE_TYPES = ['None', 'Bezier']

	def __init__(self, parent):
		#GimpIOBase.__init__(self, parent)
		self.strokeType = 1 # one of self.STROKE_TYPES
		self.closedShape = True
		self.points = []

	def decode(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		ioBuf = IO(data, index, boolSize=32)
		self.strokeType = ioBuf.u32
		self.closedShape = ioBuf.bool
		numFloatsPerPoint = ioBuf.u32
		numPoints = ioBuf.u32
		for _ in range(numPoints):
			gp = GimpPoint(self)
			ioBuf.index = gp.decode(ioBuf.data, ioBuf.index, numFloatsPerPoint)
			self.points.append(gp)
		return ioBuf.index

	def encode(self):
		"""
		encode to binary data
		"""
		ioBuf = IO(boolSize=32)
		ioBuf.u32 = self.strokeType
		ioBuf.bool = self.closedShape
		#ioBuf.u32 = numFloatsPerPoint
		#ioBuf.u32 = numPoints
		for gp in self.points:
			ioBuf.addBytes(gp.encode())
		return ioBuf.data

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Stroke Type: ' + self.STROKE_TYPES[self.strokeType])
		ret.append('Closed: ' + str(self.closedShape))
		ret.append('Points: ')
		for point in self.points:
			ret.append(point.__repr__(indent + '\t'))
		return indent + (('\n' + indent).join(ret))


class GimpPoint:
	"""
	A single point within a stroke
	"""

	POINT_TYPES = ['Anchor', 'Bezier control point']

	def __init__(self, parent):
		#GimpIOBase.__init__(self, parent)
		self.x = 0
		self.y = 0
		self.pressure = 1.0
		self.xTilt = 0.5
		self.yTilt = 0.5
		self.wheel = 0.5
		self.pointType = 0

	def decode(self, data, index=0, numFloatsPerPoint=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		:param numFloatsPerPoint: required so we know
			how many different brush dynamic measurements are
			inside each point
		"""
		ioBuf = IO(data, index, boolSize=32)
		self.pressure = 1.0
		self.xTilt = 0.5
		self.yTilt = 0.5
		self.wheel = 0.5
		self.pointType = ioBuf.u32
		if numFloatsPerPoint < 1:
			numFloatsPerPoint = (len(ioBuf.data) - ioBuf.index) / 4
		self.x = ioBuf.float
		self.y = ioBuf.float
		if numFloatsPerPoint > 2:
			self.pressure = ioBuf.float
			if numFloatsPerPoint > 3:
				self.xTilt = ioBuf.float
				if numFloatsPerPoint > 4:
					self.yTilt = ioBuf.float
					if numFloatsPerPoint > 5:
						self.wheel = ioBuf.float
		return ioBuf.index

	def encode(self):
		"""
		encode to binary data
		"""
		ioBuf = IO(boolSize=32)
		ioBuf.u32 = self.pointType
		ioBuf.float = self.x
		ioBuf.float = self.y
		if self.pressure is not None:
			ioBuf.float = self.pressure
			if self.xTilt is not None:
				ioBuf.float = self.xTilt
				if self.yTilt is not None:
					ioBuf.float = self.yTilt
					if self.wheel is not None:
						ioBuf.float = self.wheel
		return ioBuf.data

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Location: (' + str(self.x) + ',' + str(self.y) + ')')
		ret.append('Pressure: ' + str(self.pressure))
		ret.append('Location: (' + str(self.xTilt) + ',' + str(self.yTilt) + ')')
		ret.append('Wheel: ' + str(self.wheel))
		return indent + (('\n' + indent).join(ret))



#!/usr/bin/env python3
"""
Pure python implementation of the gimp gpl palette format
"""
from __future__ import annotations
import argparse

class GimpGplPalette:
	""" Pure python implementation of the gimp gpl palette format """
	def __init__(self, fileName=None):
		self.name = ''
		self.columns = 16
		self.colors = []
		self.colorNames = []
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: Union[BytesIO, str]):
		"""
		load a gimp file

		:param fileName: can be a file name or a file-like object
		"""
		if hasattr(fileName, 'read'):
			self.fileName = fileName.name
			f = fileName
		else:
			self.fileName = fileName
			f = open(fileName, 'r')
		data = f.read()
		f.close()
		self.decode(data)

	def decode(self, data):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		"""
		data = [s.strip() for s in data.split('\n')]
		if data[0] != "GIMP Palette":
			raise Exception('File format error.  Magic value mismatch.')
		self.name = data[1].split(':', 1)[-1].lstrip()
		self.columns = int(data[2].split(':', 1)[-1].lstrip())
		for line in data[3:]:
			if len(line) < 1 or line[0] == "#": # Commented Line
				continue
			line = line.split(None, 4)
			if len(line) < 3:
				continue
			self.colors.append((int(line[0]), int(line[1]), int(line[2])))
			if len(line) > 3:
				self.colorNames.append(line[3])
			else:
				self.colorNames.append(None)

	def encode(self):
		""" encode to a raw data stream """
		data = []
		data.append("GIMP Palette")
		data.append('Name: ' + str(self.name))
		data.append('Columns: ' + str(self.columns))
		data.append("#")
		for i, color in enumerate(self.colors):
			colorName = self.colorNames[i]
			line = str(color[0]).rjust(3) + ' ' + str(color[1]).rjust(3) + ' ' + str(
			color[2]).rjust(3)
			if colorName is not None:
				line = line + '\t' + colorName
			data.append(line)
		return ('\n'.join(data) + '\n').encode('utf-8')

	def save(self, tofileName=None, toExtension=None):
		""" save this gimp image to a file """
		if toExtension is None:
			if tofileName is not None:
				toExtension = tofileName.rsplit('.', 1)
				if len(toExtension) > 1:
					toExtension = toExtension[-1]
				else:
					toExtension = None
		if not hasattr(tofileName, 'write'):
			f = open(tofileName, 'wb')
		f.write(self.encode())
		f.close()

	def __repr__(self, indent=''):
		""" Get a textual representation of this object """
		ret = []
		if self.fileName is not None:
			ret.append('fileName: ' + self.fileName)
		ret.append('Name: ' + str(self.name))
		ret.append('Columns: ' + str(self.columns))
		ret.append('Colors:')
		for i, color in enumerate(self.colors):
			colorName = self.colorNames[i]
			line = '(%d,%d,%d)' % color[0], color[1], color[2]
			if colorName is not None:
				line = line + ' ' + colorName
		return '\n'.join(ret)

	def __eq__(self, other):
		""" perform a comparison """
		if other.name != self.name:
			return False
		if other.columns != self.columns:
			return False
		if len(self.colors) != len(other.colors):
			return False
		for i, c in enumerate(self.colors):
			if c != other.colors[i]:
				return False
			if self.colorNames[i] != other.colorNames[i]:
				return False
		return True

if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("GimpGplPalette.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	group = parser.add_mutually_exclusive_group()
	group.add_argument("--dump", action="store_true",
	help="dump info about this file")
	args = parser.parse_args()

	gimpGplPalette = GimpGplPalette(args.xcfdocument)

	if args.dump:
		print(gimpGplPalette)



#!/usr/bin/env python3
"""
Gimp color gradient
"""
from __future__ import annotations
import argparse


class GradientSegment:
	"""
	Single segment within a gradient
	"""

	BLEND_FUNCTIONS = [
	"linear", "curved", "sinusoidal", "spherical (increasing)", "spherical (decreasing)", "step"]
	COLOR_TYPES = ["RGB", "HSV CCW", "HSV CW"]
	ENDPOINT_COLOR_TYPES = [
	"fixed", "foreground", "foreground transparent", "background", "background transparent"]

	def __init__(self):
		self.leftPosition = 0
		self.middlePosition = 0.5
		self.rightPosition = 1.0
		self.leftColor = (0, 0, 0, 0)
		self.rightColor = (255, 255, 255, 0)
		self.blendFunc = None # one of self.BLEND_FUNCTIONS
		self.colorType = None # one of self.COLOR_TYPES
		self.leftColorType = None # one of self.ENDPOINT_COLOR_TYPES
		self.rightColorType = None # one of self.ENDPOINT_COLOR_TYPES

	def getColor(self, percent):
		"""
		given a decimal percent (1.0 = 100%) retrieve
		the appropriate color for this point in the gradient
		"""
		raise NotImplementedError()

	def decode(self, data):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		"""
		data = data.split(' ')
		if len(data) < 11 or len(data) > 15:
			raise Exception('Data table is unexpected size. ' + str(len(data)))
		self.leftPosition = float(data[0])
		self.middlePosition = float(data[1])
		self.rightPosition = float(data[2])
		self.leftColor = (float(data[3]), float(data[4]), float(data[5]), float(data[6]))
		self.rightColor = (float(data[7]), float(data[8]), float(data[9]), float(data[10]))
		if len(data) >= 12:
			self.blendFunc = int(data[11])
			if len(data) >= 13:
				self.colorType = int(data[12])
				if len(data) >= 14:
					self.leftColorType = int(data[13])
					if len(data) >= 15:
						self.rightColorType = int(data[14])

	def encode(self):
		"""
		encode this to a byte array
		"""
		ret = []
		ret.append("%06f" % self.leftPosition)
		ret.append("%06f" % self.middlePosition)
		ret.append("%06f" % self.rightPosition)
		for chan in self.leftColor:
			ret.append("%06f" % chan)
		for chan in self.rightColor:
			ret.append("%06f" % chan)
		if self.blendFunc is not None:
			ret.append("%d" % self.blendFunc)
			if self.colorType is not None:
				ret.append("%d" % self.colorType)
				if self.leftColorType is not None:
					ret.append("%d" % self.leftColorType)
					if self.rightColorType is not None:
						ret.append("%d" % self.rightColorType)
		return (' '.join(ret))

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Left Position: ' + str(self.leftPosition))
		ret.append('Middle Position: ' + str(self.middlePosition))
		ret.append('Right Position: ' + str(self.rightPosition))
		ret.append('Left Color: ' + str(self.leftColor))
		ret.append('Right Color: ' + str(self.rightColor))
		ret.append('Blend Function: ' + self.BLEND_FUNCTIONS[self.blendFunc])
		ret.append('Color Type: ' + self.COLOR_TYPES[self.colorType])
		ret.append('Left Color Type: ' + self.ENDPOINT_COLOR_TYPES[self.leftColorType])
		ret.append('Right Color Type: ' + self.ENDPOINT_COLOR_TYPES[self.rightColorType])
		return ('\n' + indent).join(ret)


class GimpGgrGradient:
	"""
	Gimp color gradient

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/ggr.txt
	"""
	def __init__(self, fileName=None):
		self.fileName = None
		self.segments = []
		self.name = ''
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: Union[BytesIO, str]):
		"""
		load a gimp file

		:param fileName: can be a file name or a file-like object
		"""
		if isinstance(fileName, str):
			self.fileName = fileName
			file = open(fileName, 'rb')
		else:
			self.fileName = fileName.name
			file = fileName
		data = file.read()
		file.close()
		self.decode(data)

	def decode(self, data):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		data = data.decode('utf-8').split('\n')
		data = [l.strip() for l in data]
		if data[0] != 'GIMP Gradient':
			raise Exception('File format error.  Magic value mismatch.')
		self.name = data[1].split(':', 1)[-1].strip()
		numSegments = int(data[2])
		for i in range(numSegments):
			gs = GradientSegment()
			gs.decode(data[i + 3])
			self.segments.append(gs)

	def encode(self):
		"""
		encode this to a byte array
		"""
		ret = ['GIMP Gradient']
		ret.append('Name: ' + self.name)
		ret.append(str(len(self.segments)))
		for segment in self.segments:
			ret.append(segment.encode())
		return ('\n'.join(ret) + '\n').encode('utf-8')

	def save(self, tofileName=None):
		"""
		save this gimp image to a file
		"""
		if not hasattr(tofileName, 'write'):
			f = open(tofileName, 'wb')
		f.write(self.encode())
		f.close()

	def getColor(self, percent):
		"""
		given a decimal percent (1.0 = 100%) retrieve
		the appropriate color for this point in the gradient
		"""
		raise NotImplementedError()

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		if self.fileName is not None:
			ret.append('fileName: ' + self.fileName)
		ret.append('Name: ' + str(self.name))
		for s in self.segments:
			ret.append(s.__repr__(indent + '\t'))
		return ('\n' + indent).join(ret)


if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("GimpGgrGradient.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	parser.add_argument("--dump", action="store_true",
	help="dump info about this file")
	args = parser.parse_args()
	gimpGgrGradient = GimpGgrGradient(args.xcfdocument)
	if args.dump:
		print(gimpGgrGradient)



"""
Represents a single layer in a gimp image
"""
from __future__ import annotations
from typing import Optional
from PIL.Image import Image
from binaryiotools import IO
from .GimpIOBase import GimpIOBase
from .GimpImageHierarchy import GimpImageHierarchy
from .GimpChannel import GimpChannel

class GimpLayer(GimpIOBase):
	"""
	Represents a single layer in a gimp image
	"""

	COLOR_MODES = [
	'RGB color without alpha', 'RGB color with alpha', 'Grayscale without alpha',
	'Grayscale with alpha', 'Indexed without alpha', 'Indexed with alpha']
	PIL_MODE_TO_LAYER_MODE = {'L': 2, 'LA': 3, 'RGB': 0, 'RGBA': 1}

	def __init__(self, parent, name: Optional[str]=None, image: Optional[Image]=None):
		GimpIOBase.__init__(self, parent)
		self.width = 0
		self.height = 0
		self.colorMode = 0
		self.name = name
		self._imageHierarchy = None
		self._imageHierarchyPtr = None
		self._mask = None
		self._maskPtr = None
		self._data = None
		if image is not None:
			self.image = image # done last as it resets some of the above defaults

	def decode(self, data: bytearray, index: int=0):
		"""
		decode a byte buffer

		Steps:
		Create a new IO buffer (array of binary values)
		Grab attributes as outlined in the spec
		List of properties
		Get the image hierarchy and mask pointers
		Return the offset

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		# Create a new IO buffer (array of binary values)
		ioBuf = IO(data, index)
		# Grab attributes as outlined in the spec
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.colorMode = ioBuf.u32 # one of self.COLOR_MODES
		self.name = ioBuf.sz754
		# List of properties
		self._propertiesDecode(ioBuf)
		# Get the image hierarchy and mask pointers
		self._imageHierarchyPtr = self._pointerDecode(ioBuf)
		self._maskPtr = self._pointerDecode(ioBuf)
		self._mask = None
		self._data = data
		# Return the offset
		return ioBuf.index

	def encode(self):
		"""
		encode to byte array

		Steps:
		Create a new IO buffer (array of binary values)
		Set attributes as outlined in the spec
		List of properties
		Set the image hierarchy and mask pointers
		Return the data

		"""
		# Create a new IO buffer (array of binary values)
		dataAreaIO = IO()
		ioBuf = IO()
		# Set attributes as outlined in the spec
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.u32 = self.colorMode
		ioBuf.sz754 = self.name
		# Layer properties
		ioBuf.addBytes(self._propertiesEncode())
		# Pointer to the image heirachy structure
		dataAreaIndex = ioBuf.index + self._POINTER_SIZE * 2
		ioBuf.addBytes(self._pointerEncode(dataAreaIndex))
		dataAreaIO.addBytes(self.imageHierarchy.encode())
		#ioBuf.addBytes(self._pointerEncode_(dataAreaIndex))
		# Pointer to the layer mask
		if self.mask is not None:
			dataAreaIO.addBytes(self.mask.encode())
		ioBuf.addBytes(self._pointerEncode(dataAreaIndex + dataAreaIO.index))
		ioBuf.addBytes(dataAreaIO)
		# Return the data
		return ioBuf.data

	@property
	def mask(self):
		"""
		Get the layer mask
		"""
		if self._mask is None and self._maskPtr is not None and self._maskPtr != 0:
			self._mask = GimpChannel(self)
			self._mask.decode(self._data, self._maskPtr)
		return self._mask

	@property
	def image(self) -> Image:
		"""
		get the layer image

		NOTE: can return None!
		"""
		if self.imageHierarchy is None:
			return None
		return self.imageHierarchy.image

	@image.setter
	def image(self, image: Image):
		"""
		set the layer image

		NOTE: resets layer width, height, and colorMode
		"""
		self.height = image.height
		self.width = image.width
		if image.mode not in self.PIL_MODE_TO_LAYER_MODE:
			raise NotImplementedError('No way of handlng PIL image mode "' + image.mode + '"')
		self.colorMode = self.PIL_MODE_TO_LAYER_MODE[image.mode]
		if not self.name and isinstance(image, str):
			# try to use a fileName as the name
			self.name = image.rsplit('\\', 1)[-1].rsplit('/', 1)[-1]
		self._imageHierarchy = GimpImageHierarchy(self)
		self._imageHierarchy.image = image

	@property
	def imageHierarchy(self) -> GimpImageHierarchy:
		"""
		Get the image hierarchy objects

		This is mainly needed for deciphering image, and therefore,
		of little use to you, the user.

		NOTE: can return None if it has been fully read into an image
		"""
		if self._imageHierarchy is None and self._imageHierarchyPtr > 0:
			self._imageHierarchy = GimpImageHierarchy(self)
			self._imageHierarchy.decode(self._data, self._imageHierarchyPtr)
		return self._imageHierarchy

	@imageHierarchy.setter
	def imageHierarchy(self, imgHierarchy):
		""" set the image hierarchy """
		self._imageHierarchy = imgHierarchy


	def _forceFullyLoaded(self):
		"""
		make sure everything is fully loaded from the file
		"""
		if self.mask is not None:
			self.mask._forceFullyLoaded()
		_ = self.image # make sure the image is loaded so we can delete the hierarchy nonsense
		#self._imageHierarchy = None
		#self._data = None

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Name: ' + str(self.name))
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		ret.append('colorMode: ' + self.COLOR_MODES[self.colorMode])
		ret.append(GimpIOBase.__repr__(self, indent))
		m = self.mask
		if m is not None:
			ret.append('Mask:')
			ret.append(m.__repr__(indent + '\t'))
		return indent + (('\n' + indent).join(ret))



#!/usr/bin/env python3
"""
Pure python implementation of the OLD gimp gpb brush format
"""
from __future__ import annotations
import argparse
from binaryiotools import IO
from .GimpGbrBrush import GimpGbrBrush
from .GimpPatPattern import GimpPatPattern


class GimpGpbBrush:
	"""
	Pure python implementation of the OLD gimp gpb brush format

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/vbr.txt
	"""
	def __init__(self, fileName: Union[BytesIO, str]):
		self.brush = GimpGbrBrush()
		self.pattern = GimpPatPattern()
		if hasattr(fileName, 'read'):
			self.fileName = fileName.name
		else:
			self.fileName = fileName

	def load(self, fileName: Union[BytesIO, str]):
		"""
		load a gimp file

		:param fileName: can be a file name or a file-like object
		"""
		if isinstance(fileName, str):
			self.fileName = fileName
			file = open(fileName, 'rb')
		else:
			self.fileName = fileName.name
			file = fileName
		data = file.read()
		file.close()
		self.decode(data)

	def decode(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		index = self.brush.decode(data, index)
		#index = self.pattern.decode(data, index)
		return index

	def encode(self):
		""" encode this object to a byte array """
		ioBuf = IO()
		ioBuf.addBytes(self.brush.encode())
		ioBuf.addBytes(self.pattern.encode())
		return ioBuf.data

	def save(self, tofileName=None):
		""" save this gimp image to a file """
		if not hasattr(tofileName, 'write'):
			f = open(tofileName, 'wb')
		f.write(self.encode())

	def __repr__(self, indent=''):
		""" Get a textual representation of this object """
		ret = []
		if self.fileName is not None:
			ret.append('fileName: ' + self.fileName)
		ret.append(self.brush.__repr__(indent + '\t'))
		ret.append(self.pattern.__repr__(indent + '\t'))
		return ('\n' + indent).join(ret)


if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("GimpGbrBrush.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	parser.add_argument("--dump", action="store_true",
	help="dump info about this file")
	args = parser.parse_args()

	gimpDocument = GimpGbrBrush(args.xcfdocument)

	if args.dump:
		print(gimpDocument)



#!/usr/bin/env python3
"""
Pure python implementation of the gimp xcf file format

Currently supports:
	Loading xcf files
	Getting image hierarchy and info
	Getting image for each layer (PIL image)
Currently not supporting:
	Saving
	Programatically alter documents (add layer, etc)
	Rendering a final, compositied image
"""
from __future__ import annotations
import argparse
import copy
from io import BytesIO
from typing import Optional, Union

import PIL.ImageGrab
from PIL import Image
from blendmodes.blend import blendLayers, BlendType
from binaryiotools import IO
from .GimpIOBase import GimpIOBase
from .GimpImageHierarchy import GimpImageHierarchy
from .GimpPrecision import Precision
from .GimpLayer import GimpLayer
from .GimpChannel import GimpChannel

class GimpDocument(GimpIOBase):
	"""
	Pure python implementation of the gimp file format

	Has a series of attributes including the following:
	self._layers = None
	self._layerPtr = []
	self._channels = []
	self._channelPtr = []
	self.version = None
	self.width = 0
	self.height = 0
	self.baseColorMode = 0
	self.precision = None # Precision object
	self._data = None

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/xcf.txt
	"""
	def __init__(self, fileName=None):
		GimpIOBase.__init__(self, self)
		self._layers = []
		self._layerPtr = []
		self._channels = []
		self._channelPtr = []
		self.version = 11 # This is the most recent version
		self.width = 0
		self.height = 0
		self.baseColorMode = 0
		self.precision = None # Precision object
		self._data = None
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: Union[BytesIO, str]):
		"""
		Load a gimp xcf and decode the file. See decode for more on this
		process

		:param fileName: can be a file name or a file-like object
		"""
		if isinstance(fileName, str):
			self.fileName = fileName
			file = open(fileName, 'rb')
		else:
			self.fileName = fileName.name
			file = fileName
		data = file.read()
		file.close()
		self.decode(data)

	def decode(self, data: bytes, index: int=0):
		"""
		decode a byte buffer

		Steps:
		Create a new IO buffer (array of binary values)
		Check that the file is a valid gimp xcf
		Grab the file version
		Grab other attributes as outlined in the spec
		Get precision data using the class and ioBuf buffer
		List of properties
		Get the layers and add the pointers to them
		Get the channels and add the pointers to them
		Return the offset

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		# Create a new IO buffer (array of binary values)
		ioBuf = IO(data, index)
		# Check that the file is a valid gimp xcf
		if ioBuf.getBytes(9) != "gimp xcf ".encode('ascii'):
			raise Exception('Not a valid GIMP file')
		# Grab the file version
		version = ioBuf.cString
		if version == 'file':
			self.version = 0
		else:
			self.version = int(version[1:])
		# Grab other attributes as outlined in the spec
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.baseColorMode = ioBuf.u32
		# Get precision data using the class and ioBuf buffer
		self.precision = Precision()
		self.precision.decode(self.version, ioBuf)
		# List of properties
		self._propertiesDecode(ioBuf)
		self._layerPtr = []
		self._layers = []
		# Get the layers and add the pointers to them
		while True:
			ptr = self._pointerDecode(ioBuf)
			if ptr == 0:
				break
			self._layerPtr.append(ptr)
			lyr = GimpLayer(self)
			lyr.decode(ioBuf.data, ptr)
			self._layers.append(lyr)
		# Get the channels and add the pointers to them
		self._channelPtr = []
		self._channels = []
		while True:
			ptr = self._pointerDecode(ioBuf)
			if ptr == 0:
				break
			self._channelPtr.append(ptr)
			chnl = GimpChannel(self)
			chnl.decode(ioBuf.data, ptr)
			self._channels.append(chnl)
		# Return the offset
		return ioBuf.index

	def encode(self):
		"""
		encode to a byte array

		Steps:
		Create a new IO buffer (array of binary values)
		The file is a valid gimp xcf
		Set the file version
		Set other attributes as outlined in the spec
		Set precision data using the class and ioBuf buffer
		List of properties
		Set the layers and add the pointers to them
		Set the channels and add the pointers to them
		Return the data

		"""
		# Create a new IO buffer (array of binary values)
		ioBuf = IO()
		# The file is a valid gimp xcf
		ioBuf.addBytes("gimp xcf ")
		# Set the file version
		ioBuf.addBytes("v{0:03d}".format(self.version) + '\0')
		# Set other attributes as outlined in the spec
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.u32 = self.baseColorMode
		# Set precision data using the class and ioBuf buffer
		if self.precision is None:
			self.precision = Precision()
		self.precision.encode(self.version, ioBuf)
		# List of properties
		ioBuf.addBytes(self._propertiesEncode())
		dataAreaIdx = ioBuf.index + self._POINTER_SIZE * (len(self.layers) + len(self._channels))
		dataAreaIO = IO()
		# Set the layers and add the pointers to them
		for lyr in self.layers:
			ioBuf.pointer = dataAreaIdx + dataAreaIO.index
			dataAreaIO.addBytes(lyr.encode())
		# Set the channels and add the pointers to them
		for channel in self._channels:
			ioBuf.pointer = dataAreaIdx + dataAreaIO.index
			dataAreaIO.addBytes(channel.encode())
		ioBuf.addBytes(dataAreaIO)
		# Return the data
		return ioBuf.data

	def _forceFullyLoaded(self):
		"""
		make sure everything is fully loaded from the file
		"""
		for lyr in self.layers:
			lyr._forceFullyLoaded()
		for chan in self._channels:
			chan._forceFullyLoaded()
		# no longer try to get the data from file
		self._layerPtr = None
		self._channelPtr = None
		self._data = None

	@property
	def layers(self):
		"""
		Decode the image's layers if necessary

		TODO: need to do the same thing with self.Channels
		"""
		if len(self._layers) == 0:
			self._layers = []
			for ptr in self._layerPtr:
				lyr = GimpLayer(self)
				lyr.decode(self._data, ptr)
				self._layers.append(lyr)
			# add a reference back to this object so it doesn't go away while array is in use
			self._layers.parent = self
			# override some internal methods so we can do more with them
			self._layers._actualDelitem_ = self._layers.__delitem__
			self._layers.__delitem__ = self.deleteLayer
			self._layers._actualSetitem_ = self._layers.__delitem__
			self._layers.__setitem__ = self.setLayer
		return self._layers

	def getLayer(self, index: int):
		"""
		return a given layer
		"""
		return self.layers[index]

	def setLayer(self, _index, _l):
		"""
		assign to a given layer
		"""
		self._forceFullyLoaded()
		self._layerPtr = None # no longer try to use the pointers to get data
		#self.layers._actualSetitem(index, l)

	def newLayer(self, name: str, image: Image.Image, index: int=-1):
		"""
		create a new layer based on a PIL image

		:param name: a name for the new layer
		:param index: where to insert the new layer (default=top)
		:return: newly created GimpLayer object
		"""
		lyr = GimpLayer(self, name, image)
		lyr.imageHierarchy = GimpImageHierarchy(self, image=image)
		self.insertLayer(lyr, index)
		return lyr

	def newLayerFromClipboard(self, name: str='pasted', index: int=-1):
		"""
		Create a new image from the system clipboard.

		:param name: a name for the new layer (default="pasted")
		:param index: where to insert the new layer (default=top)
		:return: newly created GimpLayer object

		NOTE: requires pillow PIL implementation
		NOTE: only works on OSX and Windows
		"""
		image = PIL.ImageGrab.grabclipboard()
		return self.newLayer(name, image, index)

	def addLayer(self, lyr: GimpLayer):
		"""
		append a layer object to the document

		:param layer: the new layer to append
		"""
		self.insertLayer(lyr, -1)

	def appendLayer(self, l):
		"""
		append a layer object to the document

		:param layer: the new layer to append
		"""
		self.insertLayer(l, -1)

	def insertLayer(self, lyr: GimpLayer, index: int=-1):
		"""
		insert a layer object at a specific position

		:param layer: the new layer to insert
		:param index: where to insert the new layer (default=top)
		"""
		self._layers.insert(index, lyr)

	def deleteLayer(self, index: int) -> None:
		"""
		delete a layer
		"""
		self.__delitem__(index)

	# make this class act like this class is an array of layers
	def __len__(self):
		return len(self.layers)

	def __getitem__(self, index: int):
		return self.layers[index]

	def __setitem__(self, index: int, lyr):
		self.setLayer(index, lyr)

	def __delitem__(self, index: int):
		self.deleteLayer(index)

	def __inc__(self, amt):
		self.appendLayer(amt)
		return self

	@property
	def image(self):
		"""
		get a final, compiled image
		"""
		# We need to preprocess the layers to sort them into a list of layers
		# and groups. Where a group is a list of layers

		# Example Layers [layer, layer, group, layer]
		# Example Group [Layer(Group), [layer, layer, ...]]
		layers = self.layers[:] # Copy the attribute rather than write to it
		layersOut = []
		index = 0
		while index < len(layers):
			layerOrGroup = layers[index]
			if layerOrGroup.isGroup:
				elem = [layerOrGroup, []]
				index += 1
				while layers[index].itemPath is not None:
					layerCopy = copy.deepcopy(layers[index])
					layerCopy.xOffset -= layerOrGroup.xOffset
					layerCopy.yOffset -= layerOrGroup.yOffset
					elem[1].append(layerCopy)
					layers.pop(index)
				layersOut.append(elem)
			else:
				layersOut.append(layerOrGroup)
				index += 1
		return flattenAll(layersOut, (self.width, self.height))

	def save(self, tofileName=None):
		"""
		save this gimp image to a file
		"""
		self._forceFullyLoaded()
		if tofileName is None:
			tofileName = self.fileName
		if not hasattr(tofileName, 'write'):
			file = open(tofileName, 'wb')
		file.write(self.encode())

	def saveNew(self, tofileName=None):
		"""
		save a new gimp image to a file
		"""
		if tofileName is None:
			tofileName = self.fileName
		if not hasattr(tofileName, 'write'):
			f = open(tofileName, 'wb')
		f.write(self.encode())

	def __repr__(self, indent: str=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		if self.fileName is not None:
			ret.append('fileName: ' + self.fileName)
		ret.append('Version: ' + str(self.version))
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		ret.append('Base Color Mode: ' + self.COLOR_MODES[self.baseColorMode])
		ret.append('Precision: ' + str(self.precision))
		ret.append(GimpIOBase.__repr__(self))
		if self._layerPtr:
			ret.append('Layers: ')
			for l in self.layers:
				ret.append(l.__repr__('\t'))
		if self._channelPtr:
			ret.append('Channels: ')
			for ch in self._channels:
				ret.append(ch.__repr__('\t'))
		return '\n'.join(ret)

def blendModeLookup(blendmode: int, blendLookup: dict[int, BlendType],
default: BlendType=BlendType.NORMAL):
	""" Get the blendmode from a lookup table """
	if blendmode not in blendLookup:
		print("WARNING " + str(blendmode) + " is not currently supported!")
		return default
	return blendLookup[blendmode]

def rasterImageOffset(image: Image.Image, size: tuple[int, int], offsets: tuple[int, int]=(0, 0)):
	""" Rasterise an image with offset to a given size"""
	imageOffset = Image.new("RGBA", size)
	imageOffset.paste(image.convert("RGBA"), offsets, image.convert("RGBA"))
	return imageOffset

def flattenLayerOrGroup(layerOrGroup: Union[list[GimpLayer], GimpLayer], imageDimensions: tuple[int, int],
flattenedSoFar: Optional[Image.Image]=None, ignoreHidden: bool=True) -> Image.Image:
	"""Flatten a layer or group on to an image of what has already been
	flattened
	Args:
		layerOrGroup (Layer|Group): A layer or a group of layers
		imageDimensions ((int, int)): size of the image
		flattenedSoFar (PIL.Image, optional): the image of what has already
		been flattened. Defaults to None.
		ignoreHidden (bool, optional): ignore layers that are hidden. Defaults
		to True.
	Returns:
		PIL.Image: Flattened image
	"""
	blendLookup = {0: BlendType.NORMAL, 3: BlendType.MULTIPLY,
	4: BlendType.SCREEN, 5: BlendType.OVERLAY, 6: BlendType.DIFFERENCE,
	7: BlendType.ADDITIVE, 8: BlendType.NEGATION, 9: BlendType.DARKEN,
	10: BlendType.LIGHTEN, 11: BlendType.HUE, 12: BlendType.SATURATION,
	13: BlendType.COLOUR, 14: BlendType.LUMINOSITY, 15: BlendType.DIVIDE,
	16: BlendType.COLOURDODGE, 17: BlendType.COLOURBURN,
	18: BlendType.HARDLIGHT, 19: BlendType.SOFTLIGHT, 20: BlendType.GRAINEXTRACT,
	21: BlendType.GRAINMERGE, 23: BlendType.OVERLAY, 24: BlendType.HUE,
	25: BlendType.SATURATION, 26: BlendType.COLOUR, 27: BlendType.LUMINOSITY,
	28: BlendType.NORMAL, 30: BlendType.MULTIPLY, 31: BlendType.SCREEN,
	32: BlendType.DIFFERENCE, 33: BlendType.ADDITIVE, 34: BlendType.NEGATION,
	35: BlendType.DARKEN, 36: BlendType.LIGHTEN, 37: BlendType.HUE,
	38: BlendType.SATURATION, 39: BlendType.COLOUR, 40: BlendType.LUMINOSITY,
	41: BlendType.DIVIDE, 42: BlendType.COLOURDODGE,
	43: BlendType.COLOURBURN, 44: BlendType.HARDLIGHT, 45: BlendType.SOFTLIGHT,
	46: BlendType.GRAINEXTRACT, 47: BlendType.GRAINMERGE,
	48: BlendType.VIVIDLIGHT, 49: BlendType.PINLIGHT, 52: BlendType.EXCLUSION}

	if isinstance(layerOrGroup, list):
		if ignoreHidden and not layerOrGroup[0].visible:
			foregroundRaster = Image.new("RGBA", imageDimensions)
		else: # A group is a list of layers
			# (see flattenAll)
			foregroundRaster = rasterImageOffset(flattenAll(layerOrGroup[1],
			imageDimensions, ignoreHidden), imageDimensions, (layerOrGroup[0].xOffset,
			layerOrGroup[0].yOffset))
		if flattenedSoFar is None:
			return foregroundRaster
		return blendLayers(flattenedSoFar, foregroundRaster,
		blendModeLookup(layerOrGroup[0].blendMode, blendLookup),
		layerOrGroup[0].opacity)

	if ignoreHidden and not layerOrGroup.visible:
		foregroundRaster = Image.new("RGBA", imageDimensions)
	else:
		# Get a raster image and apply blending
		foregroundRaster = rasterImageOffset(layerOrGroup.image, imageDimensions,
		(layerOrGroup.xOffset, layerOrGroup.yOffset))
	if flattenedSoFar is None:
		return foregroundRaster
	return blendLayers(flattenedSoFar, foregroundRaster,
	blendModeLookup(layerOrGroup.blendMode, blendLookup),
	layerOrGroup.opacity)


def flattenAll(layers, imageDimensions, ignoreHidden=True):
	"""Flatten a list of layers and groups

	Note the bottom layer is at the end of the list

	Args:
		layers ([Layer|Group]): A list of layers and groups
		imageDimensions ((int, int)): size of the image
		been flattened. Defaults to None.
		ignoreHidden (bool, optional): ignore layers that are hidden. Defaults
		to True.
	Returns:
		PIL.Image: Flattened image
	"""
	end = len(layers) - 1
	flattenedSoFar = flattenLayerOrGroup(layers[end], imageDimensions, ignoreHidden=ignoreHidden)
	for l in range(end - 1, -1, -1):
		flattenedSoFar = flattenLayerOrGroup(layers[l], imageDimensions,
		flattenedSoFar=flattenedSoFar, ignoreHidden=ignoreHidden)
	return flattenedSoFar



### ARGPARSE STUFF ###

if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("GimpGbrBrush.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	group = parser.add_mutually_exclusive_group()
	group.add_argument("--dump", action="store_true",
	help="dump info about this file")
	group.add_argument("--showLayer", action="store_true",
	help="show layer(s) (use * for all)")
	group.add_argument("--saveLayer", action="store_true",
	help="save layer(s) out to file")
	group.add_argument("--save", action="store",
	help="save out the brush image")
	args = parser.parse_args()

	gimpDocument = GimpDocument(args.xcfdocument)

	if args.dump:
		print(gimpDocument)
	if args.showLayer:
		if args.showLayer == '*':
			for layer in range(len(gimpDocument.layers)):
				im = gimpDocument.layers[layer].image
				showLayer(im, layer)
		else:
			im = gimpDocument.layers[int(args.showLayer)].image
			showLayer(im, int(args.showLayer))
	if args.saveLayer:
		layer = args.saveLayer.split(',', 1)
		if len(layer) > 1:
			fileName = layer[1]
		else:
			fileName = 'layer *.png'
		layer = args.saveLayer[0]
		if layer == '*':
			if fileName.find('*') < 0:
				fileName = '.'.join(fileName.split('.', 1).insert(1, '*'))
			for n in range(len(gimpDocument.layers)):
				saveLayer(gimpDocument, n, fileName)
		else:
			saveLayer(gimpDocument, int(layer), fileName)
	if args.save:
		gimpDocument.save(args.save)


def showLayer(image, l):
	""" show a layer """
	if image is None:
		print('No image for layer', l)
	else:
		print('Showing layer', l)
		image.show()


def saveLayer(gimpDoc, l, fileName: Union[BytesIO, str]):
	""" save a layer """
	iteration = gimpDoc.layers[l].image
	if iteration is None:
		print('No image for layer', l)
	else:
		fn2 = fileName.replace('*', str(l))
		print('saving layer', fn2)
		iteration.save(fn2)



"""
Represents a single channel or mask in a gimp image
"""
from __future__ import annotations
from typing import Optional
from PIL import Image

from binaryiotools import IO
from .GimpIOBase import GimpIOBase
from .GimpImageHierarchy import GimpImageHierarchy

class GimpChannel(GimpIOBase):
	"""
	Represents a single channel or mask in a gimp image
	"""
	def __init__(self, parent, name: str="", image: Optional[Image.Image]=None):
		GimpIOBase.__init__(self, parent)
		self.width = 0
		self.height = 0
		self.name = name
		self._imageHierarchy = None
		self._imageHierarchyPtr = None
		if image is not None: # this is last because image can reset values
			self.image = image
		self._data = None

	def decode(self, data: bytearray, index: int=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		ioBuf = IO(data, index)
		#print 'Decoding channel at',index
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.name = ioBuf.sz754
		self._propertiesDecode(ioBuf)
		self._imageHierarchyPtr = self._pointerDecode(ioBuf)
		self._data = ioBuf.data
		return ioBuf.index

	def encode(self):
		"""
		encode this object to a byte buffer
		"""
		ioBuf = IO()
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.sz754 = self.name
		ioBuf.addBytes(self._propertiesEncode())
		imgH = self._imageHierarchyPtr
		if imgH is None:
			imgH = 0
		ioBuf.addBytes(self._pointerEncode(imgH))
		return ioBuf.data

	@property
	def image(self) -> Optional[Image.Image]:
		"""
		get a final, compiled image
		"""
		return self.imageHierarchy.image

	@image.setter
	def image(self, image: Image.Image):
		"""
		get a final, compiled image
		"""
		self.width = image.width
		self.height = image.height
		if not self.name and isinstance(image, str):
			# try to use a fileName as the name
			self.name = image.rsplit('\\', 1)[-1].rsplit('/', 1)[-1]
		self._imageHierarchy = GimpImageHierarchy(self, image)

	def _forceFullyLoaded(self):
		"""
		make sure everything is fully loaded from the file
		"""
		_ = self.image # make sure the image is loaded so we can delete the hierarchy nonsense
		self._imageHierarchyPtr = None
		self._data = None

	@property
	def imageHierarchy(self):
		"""
		Get the image hierarchy

		This is mainly used for decoding the image, so
		not much use to you.
		"""
		if self._imageHierarchy is None:
			self._imageHierarchy = GimpImageHierarchy(self)
			self._imageHierarchy.decode(self._data, self._imageHierarchyPtr)
		return self._imageHierarchy

	def __repr__(self, indent: str='') -> str:
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Name: ' + str(self.name))
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		ret.append(GimpIOBase.__repr__(self, indent))
		return indent + (('\n' + indent).join(ret))



#!/usr/bin/env python3
"""
Pure python implementation of a gimp pattern file
"""
from __future__ import annotations
import argparse
import PIL.Image
from binaryiotools import IO


class GimpPatPattern:
	"""
	Pure python implementation of a gimp pattern file

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/pat.txt
	"""

	COLOR_MODES = [None, 'L', 'LA', 'RGB', 'RGBA']

	def __init__(self, fileName=None):
		self.fileName = None
		self.version = 1
		self.width = 0
		self.height = 0
		self.bpp = 4
		self.mode = self.COLOR_MODES[self.bpp]
		self.name = ''
		self._rawImage = None
		self._image = None
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: Union[BytesIO, str]):
		"""
		load a gimp file

		:param fileName: can be a file name or a file-like object
		"""
		if isinstance(fileName, str):
			self.fileName = fileName
			file = open(fileName, 'rb')
		else:
			self.fileName = fileName.name
			file = fileName
		data = file.read()
		file.close()
		self.decode(data)

	def decode(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		ioBuf = IO(data, index)
		headerSize = ioBuf.u32
		self.version = ioBuf.u32
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.bpp = ioBuf.u32
		self.mode = self.COLOR_MODES[self.bpp]
		magic = ioBuf.getBytes(4)
		if magic.decode('ascii') != 'GPAT':
			raise Exception('File format error.  Magic value mismatch.')
		nameLen = headerSize - ioBuf.index
		self.name = ioBuf.getBytes(nameLen).decode('UTF-8')
		self._rawImage = ioBuf.getBytes(self.width * self.height * self.bpp)
		self._image = None

	def encode(self):
		"""
		encode to a byte buffer
		"""
		ioBuf = IO()
		ioBuf.u32 = 24 + len(self.name)
		ioBuf.u32 = self.version
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.u32 = len(self.image.mode)
		ioBuf.addBytes('GPAT')
		ioBuf.addBytes(self.name.encode('utf-8'))
		if self._rawImage is None:
			rawImage = self.image.tobytes(encoder_name='raw')
		else:
			rawImage = self._rawImage
		ioBuf.addBytes(rawImage)
		return ioBuf.data

	@property
	def size(self):
		"""
		the size of the pattern
		"""
		return (self.width, self.height)

	@property
	def image(self):
		"""
		get a final, compiled image
		"""
		if self._image is None:
			if self._rawImage is None:
				return None
			self._image = PIL.Image.frombytes(self.mode,
			self.size,
			self._rawImage,
			decoder_name='raw')
		return self._image

	@image.setter
	def image(self, image):
		self._image = image
		self._rawImage = None

	def save(self, tofileName=None, toExtension=None):
		"""
		save this gimp image to a file
		"""
		asImage = False
		if toExtension is None:
			if tofileName is not None:
				toExtension = tofileName.rsplit('.', 1)
				if len(toExtension) > 1:
					toExtension = toExtension[-1]
				else:
					toExtension = None
		if toExtension is not None and toExtension != 'pat':
			asImage = True
		if asImage:
			self.image.save(tofileName)
		else:
			if not hasattr(tofileName, 'write'):
				f = open(tofileName, 'wb')
			f.write(self.encode())
			f.close()

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		if self.fileName is not None:
			ret.append('fileName: ' + self.fileName)
		ret.append('Name: ' + str(self.name))
		ret.append('Version: ' + str(self.version))
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		ret.append('BPP: ' + str(self.bpp))
		ret.append('Mode: ' + str(self.mode))
		return '\n'.join(ret)


if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("GimpPatPattern.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	group = parser.add_mutually_exclusive_group()
	group.add_argument("--dump", action="store_true",
	help="dump info about this file")
	group.add_argument("--show", action="store_true",
	help="show the image")
	group.add_argument("--save", action="store",
	help="save out the image")
	args = parser.parse_args()

	gimpPatPattern = GimpPatPattern(args.xcfdocument)

	if args.dump:
		print(gimpPatPattern)
	if args.show:
		gimpPatPattern.image.show()
	if args.save:
		gimpPatPattern.image.save(args.save)



"""
Gets packed pixels from a gimp image

NOTE: This was originally designed to be a hierarchy, like
	an image pyramid, through in practice they only use the
	top level of the pyramid (64x64) and ignore the rest.
"""
from __future__ import annotations
from typing import Optional
from PIL import Image
from .GimpIOBase import IO, GimpIOBase
from .GimpImageLevel import GimpImageLevel


class GimpImageHierarchy(GimpIOBase):
	"""
	Gets packed pixels from a gimp image

	NOTE: This was originally designed to be a hierarchy, like
		an image pyramid, through in practice they only use the
		top level of the pyramid (64x64) and ignore the rest.
	"""
	def __init__(self, parent, image: Optional[Image.Image]=None):
		GimpIOBase.__init__(self, parent)
		self.width: int = 0
		self.height: int = 0
		self.bpp: int = 0 # Number of bytes per pixel given
		self._levelPtrs = []
		self._levels = None
		self._data = None
		if image is not None: # NOTE: can override earlier parameters
			self.image = image

	def decode(self, data: bytearray, index: int=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		if not data:
			raise Exception('No data!')
		ioBuf = IO(data, index)
		#print 'Decoding channel at',index
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.bpp = ioBuf.u32
		if self.bpp < 1 or self.bpp > 4:
			msg = """'Unexpected bytes-per-pixel for image data (""" + str(self.bpp) + """).
				Probably means file corruption."""
			raise Exception(msg)
		while True:
			ptr = self._pointerDecode(ioBuf)
			if ptr == 0:
				break
			self._levelPtrs.append(ptr)
		if self._levelPtrs: # remove "dummy" level pointers
			self._levelPtrs = [self._levelPtrs[0]]
		self._data = data
		return ioBuf.index

	def encode(self):
		"""
		encode this object to a byte buffer
		"""
		dataioBuf = IO()
		ioBuf = IO()
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.u32 = self.bpp
		dataIndex = ioBuf.index + self._POINTER_SIZE * (len(self.levels) + 1)
		for level in self.levels:
			ioBuf.addBytes(self._pointerEncode(dataIndex + dataioBuf.index)) # TODO: This may be incorrect
			dataioBuf.addBytes(level.encode())
		ioBuf.addBytes(self._pointerEncode(0))
		ioBuf.addBytes(dataioBuf.data)
		return ioBuf.data

	@property
	def levels(self):
		"""
		Get the levels within this hierarchy

		Presently hierarchy is not really used by gimp,
		so this returns an array of one item
		"""
		if self._levels is None:
			for ptr in self._levelPtrs:
				level = GimpImageLevel(self)
				level.decode(self._data, ptr)
				self._levels = [level]
		return self._levels

	@property
	def image(self) -> Optional[Image.Image]:
		"""
		get a final, compiled image
		"""
		if not self.levels:
			return None
		return self.levels[0].image

	@image.setter
	def image(self, image: Image.Image):
		"""
		set the image
		"""
		self.width = image.width
		self.height = image.height
		if image.mode not in ['L', 'LA', 'RGB', 'RGBA']:
			raise NotImplementedError('Unsupported PIL image type')
		self.bpp = len(image.mode)
		self._levelPtrs = None
		#self._levels = [GimpImageLevel(self, image)]

	def __repr__(self, indent: str=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		ret.append('Bytes Per Pixel: ' + str(self.bpp))
		return indent + (('\n' + indent).join(ret))



#!/usr/bin/env python3
"""
Parasites are arbitrary (meta)data strings that can be attached to a document tree item

They are used to store things like last-used plugin settings, gamma adjuetments, etc.

Format of known parasites:
	https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/parasites.txt
"""
from __future__ import annotations
from binaryiotools import IO

#TODO: how to best use these for our puproses??
KNOWN_DOCUMENT_PARASITES = [
"jpeg-save-defaults", "png-save-defaults", "<plug-in>/_fu_data", "exif-orientation-rotate"]
KNOWN_IMAGE_PARASITES = [
"gimp-comment", "gimp-brush-name", "gimp-brush-pipe-name", "gimp-brush-pipe-parameters",
"gimp-image-grid", "gimp-pattern-name", "tiff-save-options", "jpeg-save-options",
"jpeg-exif-data", "jpeg-original-settings", "gamma", "chromaticity", "rendering-intent",
"hot-spot", "exif-data", "gimp-metadata", "icc-profile", "icc-profile-name", "decompose-data",
"print-settings", "print-page-setup", "dcm/XXXX-XXXX-AA"]
KNOWN_LAYER_PARASITES = ["gimp-text-layer", "gfig"]


class GimpParasite:
	"""
	Parasites are arbitrary (meta)data strings that can be attached to a document tree item

	They are used to store things like last-used plugin settings, gamma adjuetments, etc.

	Format of known parasites:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/parasites.txt
	"""
	def __init__(self):
		self.name = ''
		self.flags = 0
		self.data = None

	def decode(self, data: bytearray, index: int=0) -> int:
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		ioBuf = IO(data, index)
		self.name = ioBuf.sz754
		self.flags = ioBuf.u32
		dataLength = ioBuf.u32
		self.data = ioBuf.getBytes(dataLength)
		return ioBuf.index

	def encode(self):
		"""
		encode a byte buffer

		:param data: data buffer to encode
		:param index: index within the buffer to start at
		"""
		ioBuf = IO()
		ioBuf.sz754 = self.name
		ioBuf.u32 = self.flags
		ioBuf.u32 = len(self.data)
		ioBuf.addBytes(self.data)
		return ioBuf.data

	def __repr__(self, indent: str='') -> str:
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Name: ' + str(self.name))
		ret.append('Flags: ' + str(self.flags))
		ret.append('Data Len: ' + str(len(self.data)))
		return indent + (('\n' + indent).join(ret))
